# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from .comparison import ALCHEMY_TYPES, ExpressionMatcher, PrettyExpression


class AlchemyUnittestMixin(object):
    """
    Unittest class mixin for asserting that different SQLAlchemy
    expressions are the same.

    Uses SQLAlchemyExpressionMatcher to do the comparison.

    For example::

        >>> from sqlalchemy.sql.expression import column
        >>> import unittest
        >>> class FooTest(AlchemyUnittestMixin, unittest.TestCase):
        ...     def test_true(self):
        ...         c = column('column')
        ...         self.assertEqual(c == 5, c == 5)
        ...     def test_false(self):
        ...         c = column('column')
        ...         self.assertEqual(c == 5, c == 10)
        >>> FooTest('test_true').test_true()
        >>> FooTest('test_false').test_false()
        Traceback (most recent call last):
        ...
        AssertionError: BinaryExpression(sql='"column" = :column_1', params={'column_1': 5}) != BinaryExpression(sql='"column" = :column_1', params={'column_1': 10})
    """

    def __init__(self, *args, **kwargs):
        super(AlchemyUnittestMixin, self).__init__(*args, **kwargs)

        # add sqlalchemy expression type which will allow to
        # use self.assertEqual
        for t in ALCHEMY_TYPES:
            self.addTypeEqualityFunc(t, "assertSQLAlchemyExpressionEqual")

    def assertSQLAlchemyExpressionEqual(self, left, right, msg=None):
        """
        Assert that two given sqlalchemy expressions are equal
        as determined by SQLAlchemyExpressionMatcher
        """
        if ExpressionMatcher(left) != right:
            raise self.failureException(
                msg
                or "{!r} != {!r}".format(
                    PrettyExpression(left), PrettyExpression(right)
                )
            )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from functools import partial
from itertools import chain, takewhile

from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound

from .comparison import ExpressionMatcher
from .compat import mock
from .utils import (
    build_identity_map,
    copy_and_update,
    get_item_attr,
    indexof,
    raiser,
    setattr_tmp,
)


Call = type(mock.call)


class UnorderedTuple(tuple):
    """
    Same as tuple except in comparison order does not matter

    For example::

        >>> UnorderedTuple((1, 2, 3)) == (3, 2, 1)
        True
    """

    def __eq__(self, other):
        if len(self) != len(other):
            return False

        other = list(other)
        for i in self:
            try:
                other.remove(i)
            except ValueError:
                return False

        return True


class UnorderedCall(Call):
    """
    Same as Call except in comparison order of parameters does not matter

    For example::

        >>> UnorderedCall(((1, 2, 3), {'hello': 'world'})) == Call(((3, 2, 1), {'hello': 'world'}))
        True
    """

    def __eq__(self, other):
        _other = list(other)
        _other[-2] = UnorderedTuple(other[-2])
        other = Call(
            tuple(_other),
            **{k.replace("_mock_", ""): v for k, v in vars(other).items()}
        )

        return super(UnorderedCall, self).__eq__(other)


def sqlalchemy_call(call, with_name=False, base_call=Call):
    """
    Convert ``mock.call()`` into call with all parameters wrapped with ``ExpressionMatcher``

    For example::

        >>> args, kwargs = sqlalchemy_call(mock.call(5, foo='bar'))
        >>> isinstance(args[0], ExpressionMatcher)
        True
        >>> isinstance(kwargs['foo'], ExpressionMatcher)
        True
    """
    try:
        args, kwargs = call
    except ValueError:
        name, args, kwargs = call
    else:
        name = ""

    args = tuple([ExpressionMatcher(i) for i in args])
    kwargs = {k: ExpressionMatcher(v) for k, v in kwargs.items()}

    if with_name:
        return base_call((name, args, kwargs))
    else:
        return base_call((args, kwargs), two=True)


class AlchemyMagicMock(mock.MagicMock):
    """
    MagicMock for SQLAlchemy which can compare alchemys expressions in assertions

    For example::

        >>> from sqlalchemy import or_
        >>> from sqlalchemy.sql.expression import column
        >>> c = column('column')
        >>> s = AlchemyMagicMock()

        >>> _ = s.filter(or_(c == 5, c == 10))

        >>> _ = s.filter.assert_called_once_with(or_(c == 5, c == 10))
        >>> _ = s.filter.assert_any_call(or_(c == 5, c == 10))
        >>> _ = s.filter.assert_has_calls([mock.call(or_(c == 5, c == 10))])

        >>> s.reset_mock()
        >>> _ = s.filter(c == 5)
        >>> _ = s.filter.assert_called_once_with(c == 10)
        Traceback (most recent call last):
        ...
        AssertionError: Expected call: filter(BinaryExpression(sql='"column" = :column_1', params={'column_1': 10}))
        Actual call: filter(BinaryExpression(sql='"column" = :column_1', params={'column_1': 5}))
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("__name__", "Session")
        super(AlchemyMagicMock, self).__init__(*args, **kwargs)

    def _format_mock_call_signature(self, args, kwargs):
        name = self._mock_name or "mock"
        args, kwargs = sqlalchemy_call(mock.call(*args, **kwargs))
        return mock._format_call_signature(name, args, kwargs)

    def assert_called_with(self, *args, **kwargs):
        args, kwargs = sqlalchemy_call(mock.call(*args, **kwargs))
        return super(AlchemyMagicMock, self).assert_called_with(*args, **kwargs)

    def assert_any_call(self, *args, **kwargs):
        args, kwargs = sqlalchemy_call(mock.call(*args, **kwargs))
        with setattr_tmp(
            self,
            "call_args_list",
            [sqlalchemy_call(i) for i in self.call_args_list],
        ):
            return super(AlchemyMagicMock, self).assert_any_call(
                *args, **kwargs
            )

    def assert_has_calls(self, calls, any_order=False):
        calls = [sqlalchemy_call(i) for i in calls]
        with setattr_tmp(
            self,
            "mock_calls",
            type(self.mock_calls)(
                [sqlalchemy_call(i) for i in self.mock_calls]
            ),
        ):
            return super(AlchemyMagicMock, self).assert_has_calls(
                calls, any_order
            )


class UnifiedAlchemyMagicMock(AlchemyMagicMock):
    """
    MagicMock which unifies common SQLALchemy session functions for easier assertions.

    For example::

        >>> from sqlalchemy.sql.expression import column
        >>> c = column('column')

        >>> s = UnifiedAlchemyMagicMock()
        >>> s.query(None).filter(c == 'one').filter(c == 'two').all()
        []
        >>> s.query(None).filter(c == 'three').filter(c == 'four').all()
        []
        >>> s.filter.call_count
        2
        >>> s.filter.assert_any_call(c == 'one', c == 'two')
        >>> s.filter.assert_any_call(c == 'three', c == 'four')

    In addition, mock data be specified to stub real DB interactions.
    Result-sets are specified per filtering criteria so that unique data
    can be returned depending on query/filter/options criteria.
    Data is given as a list of ``(criteria, result)`` tuples where ``criteria``
    is a list of calls.
    Reason for passing data as a list vs a dict is that calls and SQLAlchemy
    expressions are not hashable hence cannot be dict keys.

    For example::

        >>> from sqlalchemy import Column, Integer, String
        >>> from sqlalchemy.ext.declarative import declarative_base

        >>> Base = declarative_base()

        >>> class SomeClass(Base):
        ...     __tablename__ = 'some_table'
        ...     pk1 = Column(Integer, primary_key=True)
        ...     pk2 = Column(Integer, primary_key=True)
        ...     name =  Column(String(50))
        ...     def __repr__(self):
        ...         return str(self.pk1)

        >>> s = UnifiedAlchemyMagicMock(data=[
        ...     (
        ...         [mock.call.query('foo'),
        ...          mock.call.filter(c == 'one', c == 'two')],
        ...         [SomeClass(pk1=1, pk2=1), SomeClass(pk1=2, pk2=2)]
        ...     ),
        ...     (
        ...         [mock.call.query('foo'),
        ...          mock.call.filter(c == 'one', c == 'two'),
        ...          mock.call.order_by(c)],
        ...         [SomeClass(pk1=2, pk2=2), SomeClass(pk1=1, pk2=1)]
        ...     ),
        ...     (
        ...         [mock.call.filter(c == 'three')],
        ...         [SomeClass(pk1=3, pk2=3)]
        ...     ),
        ... ])

        # .all()
        >>> s.query('foo').filter(c == 'one').filter(c == 'two').all()
        [1, 2]
        >>> s.query('bar').filter(c == 'one').filter(c == 'two').all()
        []
        >>> s.query('foo').filter(c == 'one').filter(c == 'two').order_by(c).all()
        [2, 1]
        >>> s.query('foo').filter(c == 'one').filter(c == 'three').order_by(c).all()
        []
        >>> s.query('foo').filter(c == 'three').all()
        [3]
        >>> s.query(None).filter(c == 'four').all()
        []

        # .iter()
        >>> list(s.query('foo').filter(c == 'two').filter(c == 'one'))
        [1, 2]

        # .count()
        >>> s.query('foo').filter(c == 'two').filter(c == 'one').count()
        2

        # .first()
        >>> s.query('foo').filter(c == 'one').filter(c == 'two').first()
        1
        >>> s.query('bar').filter(c == 'one').filter(c == 'two').first()

        # .one()
        >>> s.query('foo').filter(c == 'three').one()
        3
        >>> s.query('bar').filter(c == 'one').filter(c == 'two').one()
        Traceback (most recent call last):
        ...
        NoResultFound: No row was found for one()
        >>> s.query('foo').filter(c == 'one').filter(c == 'two').one()
        Traceback (most recent call last):
        ...
        MultipleResultsFound: Multiple rows were found for one()
        >>> s.query('bar').filter(c == 'one').filter(c == 'two').one_or_none()

        # .get()
        >>> s.query('foo').get((1, 1))
        1
        >>> s.query('foo').get((4, 4))
        >>> s.query('foo').filter(c == 'two').filter(c == 'one').get((1, 1))
        1
        >>> s.query('foo').filter(c == 'three').get((1, 1))
        1
        >>> s.query('foo').filter(c == 'three').get((4, 4))

        # dynamic session
        >>> class Model(Base):
        ...     __tablename__ = 'model_table'
        ...     pk1 = Column(Integer, primary_key=True)
        ...     name = Column(String)
        ...     def __repr__(self):
        ...         return str(self.pk1)
        >>> s = UnifiedAlchemyMagicMock()
        >>> s.add(SomeClass(pk1=1, pk2=1))
        >>> s.add_all([SomeClass(pk1=2, pk2=2)])
        >>> s.add_all([SomeClass(pk1=4, pk2=3)])
        >>> s.add_all([Model(pk1=4, name='some_name')])
        >>> s.query(SomeClass).all()
        [1, 2, 4]
        >>> s.query(SomeClass).get((1, 1))
        1
        >>> s.query(SomeClass).get((2, 2))
        2
        >>> s.query(SomeClass).get((3, 3))
        >>> s.query(SomeClass).filter(c == 'one').all()
        [1, 2, 4]
        >>> s.query(SomeClass).get((4, 3))
        4
        >>> s.query(SomeClass).get({"pk2": 3, "pk1": 4})
        4
        >>> s.query(Model).get(4)
        4

        # .delete()
        >>> s = UnifiedAlchemyMagicMock(data=[
        ...     (
        ...         [mock.call.query('foo'),
        ...          mock.call.filter(c == 'one', c == 'two')],
        ...         [SomeClass(pk1=1, pk2=1), SomeClass(pk1=2, pk2=2)]
        ...     ),
        ...     (
        ...         [mock.call.query('foo'),
        ...          mock.call.filter(c == 'one', c == 'two'),
        ...          mock.call.order_by(c)],
        ...         [SomeClass(pk1=2, pk2=2), SomeClass(pk1=1, pk2=1)]
        ...     ),
        ...     (
        ...         [mock.call.filter(c == 'three')],
        ...         [SomeClass(pk1=3, pk2=3)]
        ...     ),
        ...     (
        ...         [mock.call.query('foo'),
        ...          mock.call.filter(c == 'one', c == 'two', c == 'three')],
        ...         [SomeClass(pk1=1, pk2=1), SomeClass(pk1=2, pk2=2), SomeClass(pk1=3, pk2=3)]
        ...     ),
        ... ])

        >>> s.query('foo').filter(c == 'three').all()
        [3]
        >>> s.query('foo').all()
        []
        >>> s.query('foo').filter(c == 'three').delete()
        1
        >>> s.query('foo').filter(c == 'three').all()
        []
        >>> s.query('foo').filter(c == 'one').filter(c == 'two').all()
        [1, 2]
        >>> s.query('foo').filter(c == 'one').filter(c == 'two').filter(c == 'three').all()
        [1, 2, 3]
        >>> s = UnifiedAlchemyMagicMock()
        >>> s.add(SomeClass(pk1=1, pk2=1))
        >>> s.add_all([SomeClass(pk1=2, pk2=2)])
        >>> s.query(SomeClass).all()
        [1, 2]
        >>> s.query(SomeClass).delete()
        2
        >>> s.query(SomeClass).all()
        []
        >>> s = UnifiedAlchemyMagicMock()
        >>> s.add_all([SomeClass(pk1=2, pk2=2)])
        >>> s.query(SomeClass).delete()
        1
        >>> s.query(SomeClass).delete()
        0

    Also note that only within same query functions are unified.
    After ``.all()`` is called or query is iterated over, future queries are not unified.
    """

    boundary = {
        "all": lambda x: x,
        "__iter__": lambda x: iter(x),
        "count": lambda x: len(x),
        "first": lambda x: next(iter(x), None),
        "one": lambda x: (
            x[0]
            if len(x) == 1
            else raiser(
                MultipleResultsFound, "Multiple rows were found for one()"
            )
            if x
            else raiser(NoResultFound, "No row was found for one()")
        ),
        "one_or_none": lambda x: (
            x[0]
            if len(x) == 1
            else raiser(
                MultipleResultsFound,
                "Multiple rows were found for one_or_none()",
            )
            if x
            else None
        ),
        "get": lambda x, idmap: get_item_attr(build_identity_map(x), idmap),
    }
    unify = {
        "query": None,
        "add_columns": None,
        "join": None,
        "options": None,
        "group_by": None,
        "filter": UnorderedCall,
        "filter_by": UnorderedCall,
        "order_by": None,
        "limit": None,
        "distinct": None,
    }

    mutate = {"add", "add_all", "delete"}

    def __init__(self, *args, **kwargs):
        kwargs["_mock_default"] = kwargs.pop("default", [])
        kwargs["_mock_data"] = kwargs.pop("data", None)

        kwargs.update(
            {
                k: AlchemyMagicMock(
                    side_effect=partial(self._get_data, _mock_name=k)
                )
                for k in self.boundary
            }
        )

        kwargs.update(
            {
                k: AlchemyMagicMock(
                    return_value=self,
                    side_effect=partial(self._unify, _mock_name=k),
                )
                for k in self.unify
            }
        )

        kwargs.update(
            {
                k: AlchemyMagicMock(
                    return_value=None,
                    side_effect=partial(self._mutate_data, _mock_name=k),
                )
                for k in self.mutate
            }
        )

        super(UnifiedAlchemyMagicMock, self).__init__(*args, **kwargs)

    def _get_previous_calls(self, calls):
        return iter(
            takewhile(lambda i: i[0] not in self.boundary, reversed(calls))
        )

    def _get_previous_call(self, name, calls):
        # get all previous session calls within same session query
        previous_calls = self._get_previous_calls(calls)

        # skip last call
        next(previous_calls)

        return next(iter(filter(lambda i: i[0] == name, previous_calls)), None)

    def _unify(self, *args, **kwargs):
        _mock_name = kwargs.pop("_mock_name")
        submock = getattr(self, _mock_name)

        previous_method_call = self._get_previous_call(
            _mock_name, self.method_calls
        )
        previous_mock_call = self._get_previous_call(
            _mock_name, self.mock_calls
        )

        if previous_mock_call is None:
            return submock.return_value

        # remove immediate call from both filter mock as well as the parent mock object
        # as it was already registered in self.__call__ before this side-effect is called
        submock.call_count -= 1
        submock.call_args_list.pop()
        submock.mock_calls.pop()
        self.method_calls.pop()
        self.mock_calls.pop()

        # remove previous call since we will be inserting new call instead
        submock.call_args_list.pop()
        submock.mock_calls.pop()
        self.method_calls.pop(indexof(previous_method_call, self.method_calls))
        self.mock_calls.pop(indexof(previous_mock_call, self.mock_calls))

        name, pargs, pkwargs = previous_method_call
        args = pargs + args
        kwargs = copy_and_update(pkwargs, kwargs)

        submock.call_args = Call((args, kwargs), two=True)
        submock.call_args_list.append(Call((args, kwargs), two=True))
        submock.mock_calls.append(Call(("", args, kwargs)))

        self.method_calls.append(Call((name, args, kwargs)))
        self.mock_calls.append(Call((name, args, kwargs)))

        return submock.return_value

    def _get_data(self, *args, **kwargs):
        _mock_name = kwargs.pop("_mock_name")
        _mock_default = self._mock_default
        _mock_data = self._mock_data

        if _mock_data is not None:
            previous_calls = [
                sqlalchemy_call(
                    i, with_name=True, base_call=self.unify.get(i[0]) or Call
                )
                for i in self._get_previous_calls(self.mock_calls[:-1])
            ]
            sorted_mock_data = sorted(
                _mock_data, key=lambda x: len(x[0]), reverse=True
            )

            if _mock_name == "get":
                query_call = [c for c in previous_calls if c[0] == "query"][0]
                results = list(
                    chain(
                        *[
                            result
                            for calls, result in sorted_mock_data
                            if query_call in calls
                        ]
                    )
                )
                return self.boundary[_mock_name](results, *args, **kwargs)

            else:
                for calls, result in sorted_mock_data:
                    calls = [
                        sqlalchemy_call(
                            i,
                            with_name=True,
                            base_call=self.unify.get(i[0]) or Call,
                        )
                        for i in calls
                    ]
                    if all(c in previous_calls for c in calls):
                        return self.boundary[_mock_name](
                            result, *args, **kwargs
                        )

        return self.boundary[_mock_name](_mock_default, *args, **kwargs)

    def _mutate_data(self, *args, **kwargs):
        _mock_name = kwargs.get("_mock_name")
        _mock_data = self._mock_data = self._mock_data or []
        if _mock_name == "add":
            to_add = args[0]
            query_call = mock.call.query(type(to_add))

            mocked_data = next(
                iter(filter(lambda i: i[0] == [query_call], _mock_data)), None
            )
            if mocked_data:
                mocked_data[1].append(to_add)
            else:
                _mock_data.append(([query_call], [to_add]))

        elif _mock_name == "add_all":
            to_add = args[0]
            _kwargs = kwargs.copy()
            _kwargs["_mock_name"] = "add"

            for i in to_add:
                self._mutate_data(i, *args[1:], **_kwargs)
        # delete case
        else:
            _kwargs = kwargs.copy()
            # pretend like all is being called to get data
            _kwargs["_mock_name"] = "all"
            # a list of deleted items
            to_delete = list(self._get_data(*args, **_kwargs))
            num_deleted = len(to_delete)
            if to_delete:
                query_call = mock.call.query(type(to_delete[0]))
                mocked_data = next(
                    iter(filter(lambda i: i[0] == [query_call], _mock_data)),
                    None,
                )
                if mocked_data:
                    # remove objects based on the same instances
                    num_deleted = len(to_delete)
                    for row in to_delete:
                        mocked_data[1].remove(row)
            # we delete the data from the specific query
            del to_delete
            return num_deleted



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals


__version__ = "0.1.0"
__description__ = "SQLAlchemy mock helpers."
__author__ = "Miroslav Shubernetskiy"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from contextlib import contextmanager

import six
from sqlalchemy import inspect


def match_type(s, t):
    """
    Match string type

    For example::

        >>> assert type(match_type(b'hello', six.binary_type)) is six.binary_type
        >>> assert type(match_type(u'hello', six.text_type)) is six.text_type
        >>> assert type(match_type(b'hello', six.text_type)) is six.text_type
        >>> assert type(match_type(u'hello', six.binary_type)) is six.binary_type
    """
    if isinstance(s, t):
        return s
    if t is six.text_type:
        return s.decode("utf-8")
    else:
        return s.encode("utf-8")


def copy_and_update(target, updater):
    """
    Copy dictionary and update it all in one operation

    For example::

        >>> a = {'foo': 'bar'}
        >>> b = copy_and_update(a, {1: 2})
        >>> a is b
        False
        >>> b == {'foo': 'bar', 1: 2}
        True
    """
    result = target.copy()
    result.update(updater)
    return result


def indexof(needle, haystack):
    """
    Find an index of ``needle`` in ``haystack`` by looking for exact same item by pointer ids
    vs usual ``list.index()`` which finds by object comparison.

    For example::

        >>> a = {}
        >>> b = {}
        >>> haystack = [1, a, 2, b]
        >>> indexof(b, haystack)
        3
        >>> indexof(None, haystack)
        Traceback (most recent call last):
        ...
        ValueError: None is not in [1, {}, 2, {}]
    """
    for i, item in enumerate(haystack):
        if needle is item:
            return i
    raise ValueError("{!r} is not in {!r}".format(needle, haystack))


@contextmanager
def setattr_tmp(obj, name, value):
    """
    Utility for temporarily setting value in an object

    For example::

        >>> class Foo(object):
        ...     foo = 'foo'
        >>> print(Foo.foo)
        foo
        >>> with setattr_tmp(Foo, 'foo', 'bar'):
        ...     print(Foo.foo)
        bar
        >>> print(Foo.foo)
        foo

        >>> Foo.foo = None
        >>> with setattr_tmp(Foo, 'foo', 'bar'):
        ...     print(Foo.foo)
        None
    """
    original = getattr(obj, name)

    if original is None:
        yield
        return

    setattr(obj, name, value)
    try:
        yield
    finally:
        setattr(obj, name, original)


def raiser(exp, *args, **kwargs):
    """
    Utility for raising exceptions

    Useful in one-liners

    For example::

        >>> a = lambda x: not x and raiser(ValueError, 'error message')
        >>> _ = a(True)
        >>> _ = a(False)
        Traceback (most recent call last):
        ...
        ValueError: error message
    """
    raise exp(*args, **kwargs)


def build_identity_map(items):
    """
    Utility for building identity map from given sqlalchemy models

    For example::

        >>> from sqlalchemy import Column, Integer, String
        >>> from sqlalchemy.ext.declarative import declarative_base

        >>> Base = declarative_base()

        >>> class SomeClass(Base):
        ...     __tablename__ = 'some_table'
        ...     pk1 = Column(Integer, primary_key=True)
        ...     pk2 = Column(Integer, primary_key=True)
        ...     name =  Column(String(50))
        ...     def __repr__(self):
        ...         return str(self.pk1)

        >>> build_identity_map([SomeClass(pk1=1, pk2=2)])
        {(1, 2): 1}
    """
    idmap = {}

    for i in items:
        mapper = inspect(type(i)).mapper
        pk_keys = tuple(
            mapper.get_property_by_column(c).key for c in mapper.primary_key
        )
        pk = tuple(getattr(i, k) for k in sorted(pk_keys))
        idmap[pk] = i

    return idmap


def get_item_attr(idmap, access):
    """
    Utility for accessing dict by different key types (for get).

    For example::
        >>> idmap = {(1,): 2}
        >>> get_item_attr(idmap, 1)
        2
        >>> idmap = {(1,): 2}
        >>> get_item_attr(idmap, {"pk": 1})
        2
        >>> get_item_attr(idmap, (1,))
        2
    """
    if isinstance(access, dict):
        keys = []
        for names in sorted(access):
            keys.append(access[names])
        return idmap.get(tuple(keys))
    elif isinstance(access, int):
        return idmap.get((access,))
    else:
        return idmap.get(access)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals


try:
    from mock import mock
except ImportError:  # pragma: no cover
    try:
        import mock
    except ImportError:  # pragma: no cover
        from unittest import mock  # noqa # pragma: no cover



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mock_alchemy']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.3.22,<2.0.0', 'six>=1.15.0,<2.0.0']

setup_kwargs = {
    'name': 'mock-alchemy',
    'version': '0.1.0',
    'description': 'SQLAlchemy mock helpers.',
    'long_description': "===============\nMock SQLAlchemy\n===============\n.. image:: https://readthedocs.org/projects/mock-alchemy/badge/?version=latest\n    :target: https://mock-alchemy.readthedocs.io/en/latest/?badge=latest\n\n.. image:: https://img.shields.io/pypi/v/mock-alchemy.svg\n    :target: https://pypi.org/project/mock-alchemy/\n\n.. image:: https://img.shields.io/pypi/pyversions/mock-alchemy.svg\n    :target: https://pypi.org/project/mock-alchemy/\n\n.. image:: https://github.com/rajivsarvepalli/mock-alchemy/workflows/Tests/badge.svg\n    :target: https://github.com/rajivsarvepalli/mock-alchemy/actions?workflow=Tests\n\n.. image:: https://codecov.io/gh/rajivsarvepalli/mock-alchemy/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/rajivsarvepalli/mock-alchemy\n\nSQLAlchemy mock helpers.\n\n* Free software: MIT license\n* GitHub: https://github.com/rajivsarvepalli/mock-alchemy\n\nDocumentation\n-------------\n\nFull documentation is available at `http://mock-alchemy.rtfd.io/ <http://mock-alchemy.rtfd.io/>`__.\nOn the documentation, you should be able to select a version of your choice in order to view documentation\nof an older version if need be.\nThis README includes some basic examples, but more detailed examples are included in the documentation, especially in the `user guide <https://mock-alchemy.readthedocs.io/en/latest/user_guide/>`__.\n\nCredit\n----------\n\nThe original library (``alchemy-mock``) was created by Miroslav Shubernetskiy and Serkan Hoscai. This is a forked version due to a lack of updates\nin the original library. It appeared that the ``alchemy-mock`` project was no longer supported. Therefore, since I desired to add some basic support\nfor deleting, I created my own version of the library. Full credit goes to the original creators for starting and building this project. You can find the\noriginal package on `PyPi <https://pypi.org/project/alchemy-mock/>`__ and `Github <https://github.com/miki725/alchemy-mock>`__.\n\nInstalling\n----------\n\nYou can install ``mock-alchemy`` using pip::\n\n    $ pip install mock-alchemy\n\nWhy?\n----\n\nSQLAlchemy is awesome. Unittests are great.\nAccessing DB during tests - not so much.\nThis library provides easy way to mock SQLAlchemy's session\nin unittests while preserving ability to do sane asserts.\nNormally SQLAlchemy's expressions cannot be easily compared\nas comparison on binary expression produces yet another binary expression::\n\n    >>> type((Model.foo == 5) == (Model.bar == 5))\n    <class 'sqlalchemy.sql.elements.BinaryExpression'>\n\nBut they can be compared with this library::\n\n    >>> ExpressionMatcher(Model.foo == 5) == (Model.bar == 5)\n    False\n\nUsing\n-----\n\n``ExpressionMatcher`` can be directly used::\n\n    >>> from mock_alchemy.comparison import ExpressionMatcher\n    >>> ExpressionMatcher(Model.foo == 5) == (Model.foo == 5)\n    True\n\nAlternatively ``AlchemyMagicMock`` can be used to mock out SQLAlchemy session::\n\n    >>> from mock_alchemy.mocking import AlchemyMagicMock\n    >>> session = AlchemyMagicMock()\n    >>> session.query(Model).filter(Model.foo == 5).all()\n\n    >>> session.query.return_value.filter.assert_called_once_with(Model.foo == 5)\n\nIn real world though session can be interacted with multiple times to query some data.\nIn those cases ``UnifiedAlchemyMagicMock`` can be used which combines various calls for easier assertions::\n\n    >>> from mock_alchemy.mocking import UnifiedAlchemyMagicMock\n    >>> session = UnifiedAlchemyMagicMock()\n\n    >>> m = session.query(Model)\n    >>> q = m.filter(Model.foo == 5)\n    >>> if condition:\n    ...     q = q.filter(Model.bar > 10).all()\n    >>> data1 = q.all()\n    >>> data2 = m.filter(Model.note == 'hello world').all()\n\n    >>> session.filter.assert_has_calls([\n    ...     mock.call(Model.foo == 5, Model.bar > 10),\n    ...     mock.call(Model.note == 'hello world'),\n    ... ])\n\nAlso real-data can be stubbed by criteria::\n\n    >>> from mock_alchemy.mocking import UnifiedAlchemyMagicMock\n    >>> session = UnifiedAlchemyMagicMock(data=[\n    ...     (\n    ...         [mock.call.query(Model),\n    ...          mock.call.filter(Model.foo == 5, Model.bar > 10)],\n    ...         [Model(foo=5, bar=11)]\n    ...     ),\n    ...     (\n    ...         [mock.call.query(Model),\n    ...          mock.call.filter(Model.note == 'hello world')],\n    ...         [Model(note='hello world')]\n    ...     ),\n    ...     (\n    ...         [mock.call.query(AnotherModel),\n    ...          mock.call.filter(Model.foo == 5, Model.bar > 10)],\n    ...         [AnotherModel(foo=5, bar=17)]\n    ...     ),\n    ... ])\n    >>> session.query(Model).filter(Model.foo == 5).filter(Model.bar > 10).all()\n    [Model(foo=5, bar=11)]\n    >>> session.query(Model).filter(Model.note == 'hello world').all()\n    [Model(note='hello world')]\n    >>> session.query(AnotherModel).filter(Model.foo == 5).filter(Model.bar > 10).all()\n    [AnotherModel(foo=5, bar=17)]\n    >>> session.query(AnotherModel).filter(Model.note == 'hello world').all()\n    []\n\nThe ``UnifiedAlchemyMagicMock`` can partially fake session mutations\nsuch as ``session.add(instance)``. For example::\n\n    >>> session = UnifiedAlchemyMagicMock()\n    >>> session.add(Model(pk=1, foo='bar'))\n    >>> session.add(Model(pk=2, foo='baz'))\n    >>> session.query(Model).all()\n    [Model(foo='bar'), Model(foo='baz')]\n    >>> session.query(Model).get(1)\n    Model(foo='bar')\n    >>> session.query(Model).get(2)\n    Model(foo='baz')\n\nNote that its partially correct since if added models are filtered on,\nsession is unable to actually apply any filters so it returns everything::\n\n   >>> session.query(Model).filter(Model.foo == 'bar').all()\n   [Model(foo='bar'), Model(foo='baz')]\n\nFinally, ``UnifiedAlchemyMagicMock`` can partially fake deleting. Anything that can be\naccessed with ``all`` can also be deleted. For example::\n\n    >>> s = UnifiedAlchemyMagicMock()\n    >>> s.add(SomeClass(pk1=1, pk2=1))\n    >>> s.add_all([SomeClass(pk1=2, pk2=2)])\n    >>> s.query(SomeClass).all()\n    [1, 2]\n    >>> s.query(SomeClass).delete()\n    2\n    >>> s.query(SomeClass).all()\n    []\n\nNote the limitation for dynamic sessions remains the same. Additionally, the delete will not be propagated across\nqueries (only unified in the exact same query). As in if there are multiple queries in which the 'same'\nobject is present, this library considers them separate objects. For example::\n\n    >>> s = UnifiedAlchemyMagicMock(data=[\n    ...     (\n    ...         [mock.call.query('foo'),\n    ...          mock.call.filter(c == 'one', c == 'two')],\n    ...         [SomeClass(pk1=1, pk2=1), SomeClass(pk1=2, pk2=2)]\n    ...     ),\n    ...     (\n    ...         [mock.call.query('foo'),\n    ...          mock.call.filter(c == 'one', c == 'two'),\n    ...          mock.call.order_by(c)],\n    ...         [SomeClass(pk1=2, pk2=2), SomeClass(pk1=1, pk2=1)]\n    ...     ),\n    ...     (\n    ...         [mock.call.filter(c == 'three')],\n    ...         [SomeClass(pk1=3, pk2=3)]\n    ...     ),\n    ...     (\n    ...         [mock.call.query('foo'),\n    ...          mock.call.filter(c == 'one', c == 'two', c == 'three')],\n    ...         [SomeClass(pk1=1, pk2=1), SomeClass(pk1=2, pk2=2), SomeClass(pk1=3, pk2=3)]\n    ...     ),\n    ... ])\n\n    >>> s.query('foo').filter(c == 'three').delete()\n    1\n    >>> s.query('foo').filter(c == 'three').all()\n    []\n    >>> s.query('foo').filter(c == 'one').filter(c == 'two').filter(c == 'three').all()\n    [1, 2, 3]\n\nThe item referred to by :code:`c == 'three'` is still present in the filtered query despite the individual item being deleted.\n",
    'author': 'Rajiv Sarvepalli',
    'author_email': 'rajiv@sarvepalli.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rajivsarvepalli/mock-alchemy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import six
from six.moves.collections_abc import Mapping
from sqlalchemy import func
from sqlalchemy.sql.expression import column, or_

from .compat import mock
from .utils import match_type


ALCHEMY_UNARY_EXPRESSION_TYPE = type(column("").asc())
ALCHEMY_BINARY_EXPRESSION_TYPE = type(column("") == "")
ALCHEMY_BOOLEAN_CLAUSE_LIST = type(or_(column("") == "", column("").is_(None)))
ALCHEMY_FUNC_TYPE = type(func.dummy(column("")))
ALCHEMY_LABEL_TYPE = type(column("").label(""))
ALCHEMY_TYPES = (
    ALCHEMY_UNARY_EXPRESSION_TYPE,
    ALCHEMY_BINARY_EXPRESSION_TYPE,
    ALCHEMY_BOOLEAN_CLAUSE_LIST,
    ALCHEMY_FUNC_TYPE,
    ALCHEMY_LABEL_TYPE,
)


class PrettyExpression(object):
    """
    Wrapper around given expression with pretty representations

    For example::

        >>> c = column('column')
        >>> PrettyExpression(c == 5)
        BinaryExpression(sql='"column" = :column_1', params={'column_1': 5})
        >>> PrettyExpression(10)
        10
        >>> PrettyExpression(PrettyExpression(15))
        15
    """

    __slots__ = ["expr"]

    def __init__(self, e):
        if isinstance(e, PrettyExpression):
            e = e.expr
        self.expr = e

    def __repr__(self):
        if not isinstance(self.expr, ALCHEMY_TYPES):
            return repr(self.expr)

        compiled = self.expr.compile()

        return "{}(sql={!r}, params={!r})".format(
            self.expr.__class__.__name__,
            match_type(six.text_type(compiled).replace("\n", " "), str),
            {match_type(k, str): v for k, v in compiled.params.items()},
        )


class ExpressionMatcher(PrettyExpression):
    """
    Matcher for comparing SQLAlchemy expressions

    Similar to http://www.voidspace.org.uk/python/mock/examples.html#more-complex-argument-matching

    For example::

        >>> c = column('column')
        >>> c2 = column('column2')
        >>> l1 = c.label('foo')
        >>> l2 = c.label('foo')
        >>> l3 = c.label('bar')
        >>> l4 = c2.label('foo')
        >>> e1 = c.in_(['foo', 'bar'])
        >>> e2 = c.in_(['foo', 'bar'])
        >>> e3 = c.in_(['cat', 'dog'])
        >>> e4 = c == 'foo'
        >>> e5 = func.lower(c)

        >>> ExpressionMatcher(e1) == mock.ANY
        True
        >>> ExpressionMatcher(e1) == 5
        False
        >>> ExpressionMatcher(e1) == e2
        True
        >>> ExpressionMatcher(e1) != e2
        False
        >>> ExpressionMatcher(e1) == e3
        False
        >>> ExpressionMatcher(e1) == e4
        False
        >>> ExpressionMatcher(e5) == func.lower(c)
        True
        >>> ExpressionMatcher(e5) == func.upper(c)
        False
        >>> ExpressionMatcher(e1) == ExpressionMatcher(e2)
        True
        >>> ExpressionMatcher(c) == l1
        False
        >>> ExpressionMatcher(l1) == l2
        True
        >>> ExpressionMatcher(l1) == l3
        True
        >>> ExpressionMatcher(l1) == l4
        False

    It also works with nested structures::

        >>> ExpressionMatcher([c == 'foo']) == [c == 'foo']
        True
        >>> ExpressionMatcher({'foo': c == 'foo', 'bar': 5, 'hello': 'world'}) == {'foo': c == 'foo', 'bar': 5, 'hello': 'world'}
        True
    """

    def __eq__(self, other):
        if isinstance(other, type(self)):
            other = other.expr

        # if the right hand side is mock.ANY,
        # mocks comparison will not be used hence
        # we hard-code comparison here
        if isinstance(self.expr, type(mock.ANY)) or isinstance(
            other, type(mock.ANY)
        ):
            return True

        # handle string comparison bytes vs unicode in dict keys
        if isinstance(self.expr, six.string_types) and isinstance(
            other, six.string_types
        ):
            other = match_type(other, type(self.expr))

        # compare sqlalchemy public api attributes
        if type(self.expr) is not type(other):
            return False

        if not isinstance(self.expr, ALCHEMY_TYPES):

            def _(v):
                return type(self)(v)

            if isinstance(self.expr, (list, tuple)):
                return all(
                    _(i) == j
                    for i, j in six.moves.zip_longest(self.expr, other)
                )

            elif isinstance(self.expr, Mapping):
                same_keys = self.expr.keys() == other.keys()
                return same_keys and all(
                    _(self.expr[k]) == other[k] for k in self.expr.keys()
                )

            else:
                return self.expr is other or self.expr == other

        expr_compiled = self.expr.compile()
        other_compiled = other.compile()

        if six.text_type(expr_compiled) != six.text_type(other_compiled):
            return False

        if expr_compiled.params != other_compiled.params:
            return False

        return True

    def __ne__(self, other):
        return not (self == other)
